
// ad++ subscripting reduced to canonical form

#include "ad++.h"

const int N = 8 ;

int main(int argc, char** argv) {
  AdlibInit(argc, argv) ;

  Procs1 p(4) ;
  ON(p) {
    CyclicRange x(N, p.dim(0)) ;
    Range z(N) ;

    // Initialization (unoptimized)

    Array2<float> a(z, x) ;
    Array1<float> b(z) ;  // buffer

    Index m(x) ;
    OVERALL(m) {
      for(int i = 0 ; i < N ; i++)
        if (m >= i)
          a(z(i), m) = i + 1 ;
        else
          a(z(i), m) = m + 1 ;
    } ALLOVER(i) ;

    gprintf("initial a:\n") ;
    aprintf("%7.3f%*N", a, N) ;

    // Do decomposition

    float* a_dat = a.dat() ;
    Stride a0 = a.str(0), a1 = a.str(1) ;

    float* b_dat = b.dat() ;
    Stride b0 = b.str(0) ;

    for(int k = 0 ; k < N - 1 ; k++) {

      Location l(x, k) ;
      AT(l) {
        const float d = sqrt(a_dat [a0.offset(z(k)) + a1.offset(l)]) ;

        a_dat [a0.offset(z(k)) + a1.offset(l)] = d ;
        for(int s = k + 1 ; s < N ; s++)
          a_dat [a0.offset(z(s)) + a1.offset(l)] /= d ;
      } TA(l) ;

      remap(b.sect(z.triplet(k + 1, N - 1)),
            a.sect(z.triplet(k + 1, N - 1), x(k))) ;

      Index m(x.triplet(k + 1, N - 1)) ;
      OVERALL(m) {
        // NB: `x.idx(m)' is currently expressed as k + 1 + (int) m

        for(int i = k + 1 + (int) m ; i < N ; i++)
          a_dat [a0.offset(z(i)) + a1.offset(m)] -=
              b_dat [b0.offset(z(i))] * b_dat [b0.offset(z(k + 1 + (int) m))] ;
      } ALLOVER(m) ;

      gprintf("after iteration %d:\n", k) ;
      aprintf("%7.3f%*N", a, N) ;
    }

    Location l(x, N - 1) ;
    AT(l) {
      a_dat [a0.offset(z(N - 1)) + a1.offset(l)] = 
          sqrt(a_dat [a0.offset(z(N - 1)) + a1.offset(l)]) ;
    } TA(l) ;

    gprintf("final a:\n") ;
    aprintf("%7.3f%*N", a, N) ;

  } NO(p) ;

  AdlibFinalize() ;
}

